/*
 * Fixed-point format: 0xABBBBBBB
 * A == whole part      (sign + 3 bits)
 * B == fractional part (28 bits)
 *
 * Values are signed two's complement, so the effective range is:
 * 0x80000000 to 0x7fffffff
 *       -8.0 to +7.9999999962747097015380859375
 *
 * The smallest representable value is:
 * 0x00000001 == 0.0000000037252902984619140625 (i.e. about 3.725e-9)
 *
 * 28 bits of fractional accuracy represent about
 * 8.6 digits of decimal accuracy.
 *
 * Fixed-point numbers can be added or subtracted as normal
 * integers, but multiplication requires shifting the 64-bit result
 * from 56 fractional bits back to 28 (and rounding.)
 */

#ifndef FIXED_H
#define FIXED_H

typedef signed int fixed_t;

#define F_WHOLEBITS     4
#define F_FRACBITS	28

#define F_ONE		(1L << F_FRACBITS)

#define f_tofixed(x)	((fixed_t) ((x) >= 0 ? (x) * (double)F_ONE + 0.5 : (x) * (double)F_ONE - 0.5))
#define f_todouble(x)	((double)  ((x) / (double)F_ONE))

#define f_intpart(x)	((x) >> F_FRACBITS)
#define f_fracpart(x)	((x) & ((1L << F_FRACBITS) - 1))
				/* (x should be positive) */

#define f_fromint(x)	((x) << F_FRACBITS)

#define f_add(x, y)	((x) + (y))
#define f_sub(x, y)	((x) - (y))

#define f_abs(x)        (((x) < 0) ? -(x) : (x))

/*
 * This ARM V4 version is as accurate as FPM_64BIT but much faster. The
 * least significant bit is properly rounded at no CPU cycle cost!
 */
typedef   signed int fixed64hi_t;
typedef unsigned int fixed64lo_t;

#define f_mul(dest,x,y)                    \
    {   register fixed64hi_t hi;           \
        register fixed64lo_t lo;           \
	__asm                              \
	{                                  \
            smull lo, hi, x, y             \
        }	                           \
        asm("movs lo, lo, lsr #28;         \
	     adc  lo, lo, hi, lsl #4");    \
        dest = lo;                         \
    }

#define f_mula(dest,x,y)                   \
    {   register fixed64hi_t hi;           \
        register fixed64lo_t lo;           \
	__asm                              \
	{                                  \
            smull lo, hi, x, y             \
        }	                           \
        asm("movs lo, lo, lsr #28;         \
	     adc  lo, lo, hi, lsl #4");    \
        dest += lo;                        \
    }

#endif
