#include "mpc.h"
#include "mpcdec.h"
#include "MpcRequant.h"
#include "MpcHuffman.h"
#include <string.h>

/*
 *  This is a simple random number generator with good quality for audio purposes.
 *  It consists of two polycounters with opposite rotation direction and different
 *  periods. The periods are coprime, so the total period is the product of both.
 *
 *     -------------------------------------------------------------------------------------------------
 * +-> |31:30:29:28:27:26:25:24:23:22:21:20:19:18:17:16:15:14:13:12:11:10: 9: 8: 7: 6: 5: 4: 3: 2: 1: 0|
 * |   -------------------------------------------------------------------------------------------------
 * |                                                                          |  |  |  |     |        |
 * |                                                                          +--+--+--+-XOR-+--------+
 * |                                                                                      |
 * +--------------------------------------------------------------------------------------+
 *
 *     -------------------------------------------------------------------------------------------------
 *     |31:30:29:28:27:26:25:24:23:22:21:20:19:18:17:16:15:14:13:12:11:10: 9: 8: 7: 6: 5: 4: 3: 2: 1: 0| <-+
 *     -------------------------------------------------------------------------------------------------   |
 *       |  |           |  |                                                                               |
 *       +--+----XOR----+--+                                                                               |
 *                |                                                                                        |
 *                +----------------------------------------------------------------------------------------+
 *
 *
 *  The first has an period of 3*5*17*257*65537, the second of 7*47*73*178481,
 *  which gives a period of 18.410.713.077.675.721.215. The result is the
 *  XORed values of both generators.
 */

const unsigned char Parity [256] = {  // parity
    0,1,1,0,1,0,0,1,1,0,0,1,0,1,1,0,1,0,0,1,0,1,1,0,0,1,1,0,1,0,0,1,
    1,0,0,1,0,1,1,0,0,1,1,0,1,0,0,1,0,1,1,0,1,0,0,1,1,0,0,1,0,1,1,0,
    1,0,0,1,0,1,1,0,0,1,1,0,1,0,0,1,0,1,1,0,1,0,0,1,1,0,0,1,0,1,1,0,
    0,1,1,0,1,0,0,1,1,0,0,1,0,1,1,0,1,0,0,1,0,1,1,0,0,1,1,0,1,0,0,1,
    1,0,0,1,0,1,1,0,0,1,1,0,1,0,0,1,0,1,1,0,1,0,0,1,1,0,0,1,0,1,1,0,
    0,1,1,0,1,0,0,1,1,0,0,1,0,1,1,0,1,0,0,1,0,1,1,0,0,1,1,0,1,0,0,1,
    0,1,1,0,1,0,0,1,1,0,0,1,0,1,1,0,1,0,0,1,0,1,1,0,0,1,1,0,1,0,0,1,
    1,0,0,1,0,1,1,0,0,1,1,0,1,0,0,1,0,1,1,0,1,0,0,1,1,0,0,1,0,1,1,0
};

unsigned long __r1 = 1;
unsigned long __r2 = 1;

void
set_seed ( unsigned long  seed )
{
    __r1 = seed != 0 ? seed : 1;                // 0 is a forbidden value which locks the generator
    __r2 = 1;                                   // the are several (8389119) forbidden codes. 1 is not one of them
}


unsigned long
random_int ( void )
{
    unsigned long t1, t2, t3, t4;

    t3   = t1 = __r1;   t4   = t2 = __r2;       // Parity calculation is done via table lookup, this is also available
    t1  &= 0xF5;        t2 >>= 25;              // on CPUs without parity, can be implemented in C and avoid unpredictable
    t1   = Parity [t1]; t2  &= 0x63;            // jumps and slow rotate through the carry flag operations.
    t1 <<= 31;          t2   = Parity [t2];

    return (__r1 = (t3 >> 1) | t1 ) ^ (__r2 = (t4 << 1) | t2 );
}

/****************************************** SV 6 ******************************************/
void Lese_Bitstrom_SV6(mpc_params *mpp, Bit_stream_struc *bs)
{
	int n,k;
	const HuffmanTyp *Table;
	int Max_used_Band = -1;

	//reading Frameheader
	for (n=0; n<=mpp->Max_Band; ++n)
	{
		if      (n<11)           { Table = Region_A; k = 14; }
		else if (n>=11 && n<=22) { Table = Region_B; k = 7; }
		else /*if (n>=23)*/      { Table = Region_C; k = 3; }

		if (mpp->Bitrate<=128) mpp->Res[0][n] = Q_res[n][Huffman_Decode_fastest(bs, Table, k)];
		else		       mpp->Res[0][n] = Q_res[n][get_bits(bs, Q_bit[n])];

		//nur Lesen, falls kein IS
		if (!(mpp->IS_Flag && n>=mpp->Min_Band))
		{
			if (mpp->MS_used)	  mpp->MS_Flag[n] = get_bits(bs, 1);
			else                      mpp->MS_Flag[n] = 0;
			if (mpp->Bitrate<=128)    mpp->Res[1][n] = Q_res[n][Huffman_Decode_fastest(bs, Table, k)];
			else			  mpp->Res[1][n] = Q_res[n][get_bits(bs, Q_bit[n])];
		}
		else
		{
			mpp->MS_Flag[n] = 0;
			mpp->Res[1][n]=0;
		}
		//fuehre nachfolgende Operationen nur bis zum maximal enthaltenen Subband aus
		if (mpp->Res[0][n] || mpp->Res[1][n]) Max_used_Band = n;
	}

	//reading SCFI-Bundle
	for (n=0; n<=Max_used_Band; ++n)
	{
		if (mpp->Res[0][n])
			mpp->SCFI[0][n] = SCFI_Bundle_read(bs);
		if (mpp->Res[1][n] || (mpp->Res[0][n] && n>=mpp->Min_Band))
			mpp->SCFI[1][n] = SCFI_Bundle_read(bs);
	}

	//reading SCF
	Table = DSCF_Entropie;
	for (n=0; n<=Max_used_Band; ++n)
	{
		if (mpp->Res[0][n])
		{
			k = mpp->SCFI[0][n];

			//mit DSCF
			if (k & 1)
			{
				switch (k >> 1) {
				case 3:
					k = mpp->SCF_Index[0][n][0] = mpp->SCF_Index[0][n][2] + Huffman_Decode_fastest(bs, Table, 6);
					mpp->SCF_Index[0][n][1] = k;
					mpp->SCF_Index[0][n][2] = k;
				        break;
				case 1:
					k = mpp->SCF_Index[0][n][0] = mpp->SCF_Index[0][n][2] + Huffman_Decode_fastest(bs, Table, 6);
					k = mpp->SCF_Index[0][n][1] = k + Huffman_Decode_fastest(bs, Table, 6);
					mpp->SCF_Index[0][n][2] = k;
				        break;
				case 2:
					k = mpp->SCF_Index[0][n][0] = mpp->SCF_Index[0][n][2] + Huffman_Decode_fastest(bs, Table, 6);
					mpp->SCF_Index[0][n][1] = k;
					mpp->SCF_Index[0][n][2] = k + Huffman_Decode_fastest(bs, Table, 6);
				        break;
				default: /* SCFI_L[n]==0 */
					k = mpp->SCF_Index[0][n][0] = mpp->SCF_Index[0][n][2] + Huffman_Decode_fastest(bs, Table, 6);
					k = mpp->SCF_Index[0][n][1] = k + Huffman_Decode_fastest(bs, Table, 6);
					mpp->SCF_Index[0][n][2] = k + Huffman_Decode_fastest(bs, Table, 6);
				        break;
				}
			}
			//ohne DSCF
			else /*if (DSCF_Flag_L[n]==0)*/
			{
				switch (k >> 1) {
				case 3:
					k = mpp->SCF_Index[0][n][0] = get_bits(bs, 6);
					mpp->SCF_Index[0][n][1] = k;
					mpp->SCF_Index[0][n][2] = k;
				        break;
				case 1:
					mpp->SCF_Index[0][n][0] = get_bits(bs, 6);
					k = mpp->SCF_Index[0][n][1] = get_bits(bs, 6);
					mpp->SCF_Index[0][n][2] = k;
				        break;
				case 2:
					k = mpp->SCF_Index[0][n][0] = get_bits(bs, 6);
					mpp->SCF_Index[0][n][1] = k;
					mpp->SCF_Index[0][n][2] = get_bits(bs, 6);
				        break;
				default: /* SCFI_L[n]==0 */
					mpp->SCF_Index[0][n][0] = get_bits(bs, 6);
					mpp->SCF_Index[0][n][1] = get_bits(bs, 6);
					mpp->SCF_Index[0][n][2] = get_bits(bs, 6);
				        break;
				}
			}
		}
		if (mpp->Res[1][n] || (mpp->Res[0][n] && n>=mpp->Min_Band))
		{

			k = mpp->SCFI[1][n];

			//mit DSCF
			if (k & 1)
			{
				switch (k >> 1) {
				case 3:
					k = mpp->SCF_Index[1][n][0] = mpp->SCF_Index[1][n][2] + Huffman_Decode_fastest(bs, Table, 6);
					mpp->SCF_Index[1][n][1] = k;
					mpp->SCF_Index[1][n][2] = k;
				        break;
				case 1:
					k = mpp->SCF_Index[1][n][0] = mpp->SCF_Index[1][n][2] + Huffman_Decode_fastest(bs, Table, 6);
					k = mpp->SCF_Index[1][n][1] = k + Huffman_Decode_fastest(bs, Table, 6);
					mpp->SCF_Index[1][n][2] = k;
				        break;
				case 2:
					k = mpp->SCF_Index[1][n][0] = mpp->SCF_Index[1][n][2] + Huffman_Decode_fastest(bs, Table, 6);
					mpp->SCF_Index[1][n][1] = k;
					mpp->SCF_Index[1][n][2] = k + Huffman_Decode_fastest(bs, Table, 6);
				        break;
				default: /* SCFI_R[n]==0 */
					k = mpp->SCF_Index[1][n][0] = mpp->SCF_Index[1][n][2] + Huffman_Decode_fastest(bs, Table, 6);
					k = mpp->SCF_Index[1][n][1] = k + Huffman_Decode_fastest(bs, Table, 6);
					mpp->SCF_Index[1][n][2] = k + Huffman_Decode_fastest(bs, Table, 6);
				        break;
				}
			}
			//ohne DSCF
			else /*if (DSCF_Flag_R[n]==0)*/
			{
				switch (k >> 1) {
				case 3:
					k = mpp->SCF_Index[1][n][0] = get_bits(bs, 6);
					mpp->SCF_Index[1][n][1] = k;
					mpp->SCF_Index[1][n][2] = k;
				        break;
				case 1:
					mpp->SCF_Index[1][n][0] = get_bits(bs, 6);
					k = mpp->SCF_Index[1][n][1] = get_bits(bs, 6);
					mpp->SCF_Index[1][n][2] = k;
				        break;
				case 2:
					k = mpp->SCF_Index[1][n][0] = get_bits(bs, 6);
					mpp->SCF_Index[1][n][1] = k;
					mpp->SCF_Index[1][n][2] = get_bits(bs, 6);
				        break;
				default: /* SCFI_R[n]==0 */
					mpp->SCF_Index[1][n][0] = get_bits(bs, 6);
					mpp->SCF_Index[1][n][1] = get_bits(bs, 6);
					mpp->SCF_Index[1][n][2] = get_bits(bs, 6);
				        break;
				}
			}
		}
	}

	//reading Samples
	for (n=0; n<=Max_used_Band; ++n)
	{
		const HuffmanTyp *x1 = Entropie[mpp->Res[0][n]],
		                 *x2 = Entropie[mpp->Res[1][n]];
                int len1 = Entropie_Len[mpp->Res[0][n]];
                int len2 = Entropie_Len[mpp->Res[1][n]];

		for (k=0; k<36; ++k)
		{
			if (x1) mpp->Q[n][0][k] = Huffman_Decode_fastest(bs, x1, len1);
			if (x2) mpp->Q[n][1][k] = Huffman_Decode_fastest(bs, x2, len2);
		}

		if (mpp->Res[0][n]>7 || mpp->Res[1][n]>7)
		for (k=0; k<36; ++k)
		{
			if (mpp->Res[0][n]>7)
                                mpp->Q[n][0][k] = get_bits(bs, mpp->Res[0][n] - 1) - Dc[mpp->Res[0][n]];
			if (mpp->Res[1][n]>7)
                                mpp->Q[n][1][k] = get_bits(bs, mpp->Res[1][n] - 1) - Dc[mpp->Res[1][n]];
		}
	}
}

/****************************************** SV 7 ******************************************/

const signed char idx30[] = { -1, 0,+1,-1, 0,+1,-1, 0,+1,-1, 0,+1,-1, 0,+1,-1, 0,+1,-1, 0,+1,-1, 0,+1,-1, 0,+1 };
const signed char idx31[] = { -1,-1,-1, 0, 0, 0,+1,+1,+1,-1,-1,-1, 0, 0, 0,+1,+1,+1,-1,-1,-1, 0, 0, 0,+1,+1,+1 };
const signed char idx32[] = { -1,-1,-1,-1,-1,-1,-1,-1,-1, 0, 0, 0, 0, 0, 0, 0, 0, 0,+1,+1,+1,+1,+1,+1,+1,+1,+1 };
const signed char idx50[] = { -2,-1, 0,+1,+2,-2,-1, 0,+1,+2,-2,-1, 0,+1,+2,-2,-1, 0,+1,+2,-2,-1, 0,+1,+2 };
const signed char idx51[] = { -2,-2,-2,-2,-2,-1,-1,-1,-1,-1, 0, 0, 0, 0, 0,+1,+1,+1,+1,+1,+2,+2,+2,+2,+2 };

void Lese_Bitstrom_SV7(mpc_params *mpp, Bit_stream_struc *bs)
{
        int n,ch;
        int Max_used_Band = -1;
        const HuffmanTyp *Table;
        signed short *Q;

	/***************************** Header *****************************/
	Table = HuffHdr;
	//first subband
	mpp->Res[0][0] = get_bits(bs, 4);
	mpp->Res[1][0] = get_bits(bs, 4);
	if (mpp->MS_used && (mpp->Res[0][0] || mpp->Res[1][0]))
	    mpp->MS_Flag[0] = get_bits(bs, 1);
	else
	    mpp->MS_Flag[0] = 0;
	if (mpp->Res[0][0] || mpp->Res[1][0]) Max_used_Band = 0;

	//consecutive subbands
	for (n=1; n<=mpp->Max_Band; ++n)
	{
		signed int diff;

		diff = Huffman_Decode_fastest(bs, Table, 9);
		if (diff!=4) mpp->Res[0][n] = mpp->Res[0][n-1] + diff;
		else         mpp->Res[0][n] = get_bits(bs, 4);

		//nur Lesen, falls kein IS
		if (n<mpp->Min_Band)
		{
			diff = Huffman_Decode_fastest(bs, Table, 9);
			if (diff!=4) mpp->Res[1][n] = mpp->Res[1][n-1] + diff;
			else         mpp->Res[1][n] = get_bits(bs, 4);
			if (mpp->MS_used && (mpp->Res[0][n] || mpp->Res[1][n]))
                            switch ( mpp->Res[1][n] ) {
                            case -3: mpp->MS_Flag[n] = get_bits(bs, 2) << 2;
                                break;
                            case -2: mpp->MS_Flag[n] = get_bits(bs, 4);
                                break;
                            default: mpp->MS_Flag[n] = get_bits(bs, 1);
                                break;
                            }
			else
		            mpp->MS_Flag[n] = 0;
		}
		else
		{
			if(mpp->Res[0][n])
                            mpp->Res[1][n] = 18;
                        else
                            mpp->Res[1][n] = 0;
			mpp->MS_Flag[n] = 0;
		}

		//fuehre nachfolgende Operationen nur bis zum maximal enthaltenen Subband aus
		if (mpp->Res[0][n] || mpp->Res[1][n]) Max_used_Band = n;
	}

	/****************************** SCFI ******************************/
	for (n=0; n<=Max_used_Band; ++n)
	{
		if ((mpp->Res[0][n] > 0) || (mpp->Res[0][n] == -1))
			mpp->SCFI[0][n] = Huffman_Decode_SCFI(bs);
		if ((mpp->Res[1][n] > 0) || (mpp->Res[1][n] == -1))
			mpp->SCFI[1][n] = Huffman_Decode_SCFI(bs);
	}

	/**************************** SCF/DSCF ****************************/
	Table = HuffDSCF;
	for (n=0; n<=Max_used_Band; ++n)
	{
            signed int diff, k;

            for(ch = 0; ch < 2; ch++)
		if ((mpp->Res[ch][n] > 0) || (mpp->Res[ch][n] == -1))
		{
                        signed char * SCF_Index = mpp->SCF_Index[ch][n];

			switch(mpp->SCFI[ch][n]) {
			case 0:
				diff = Huffman_Decode_fastest(bs, Table, 6);
				if (diff!=8) k = SCF_Index[0] = SCF_Index[2] + diff;
				else	     k = SCF_Index[0] = get_bits(bs, 6);
				diff = Huffman_Decode_fastest(bs, Table, 6);
				if (diff!=8) k = SCF_Index[1] = k + diff;
				else	     k = SCF_Index[1] = get_bits(bs, 6);
				diff = Huffman_Decode_fastest(bs, Table, 6);
				if (diff!=8) SCF_Index[2] = k + diff;
				else	     SCF_Index[2] = get_bits(bs, 6);
				break;
			case 1:
				diff = Huffman_Decode_fastest(bs, Table, 6);
				if (diff!=8) k = SCF_Index[0] = SCF_Index[2] + diff;
				else	     k = SCF_Index[0] = get_bits(bs, 6);
				diff = Huffman_Decode_fastest(bs, Table, 6);
				if (diff!=8) k = SCF_Index[1] = k + diff;
				else         k = SCF_Index[1] = get_bits(bs, 6);
				SCF_Index[2] = k;
				break;
			case 2:
				diff = Huffman_Decode_fastest(bs, Table, 6);
				if (diff!=8) k = SCF_Index[0] = SCF_Index[2] + diff;
				else	     k = SCF_Index[0] = get_bits(bs, 6);
				SCF_Index[1] = k;
				diff = Huffman_Decode_fastest(bs, Table, 6);
				if (diff!=8) SCF_Index[2] = k + diff;
				else	     SCF_Index[2] = get_bits(bs, 6);
				break;
			default: /* SCFI[ch][n]==3 */
				diff = Huffman_Decode_fastest(bs, Table, 6);
				if (diff!=8) k = SCF_Index[0] = SCF_Index[2] + diff;
				else	     k = SCF_Index[0] = get_bits(bs, 6);
				SCF_Index[1] = k;
				SCF_Index[2] = k;
				break;
			}
		}
	}

	/***************************** Samples ****************************/
	Q = mpp->Q[0][0];
	for (n=0; n<=Max_used_Band; ++n)
	{
            int k, bits;

            for(ch = 0; ch < 2; ch++)
		switch(mpp->Res[ch][n]) {
                case  -2: case  -3:
                        for (k=0; k<36; k++) *Q++ = 0;
                        break;
		case -1:
                        for (k=0; k<36; k++ ) {
                                unsigned long tmp  = random_int ();
                                *Q++ = ((tmp >> 24) & 0xFF) + ((tmp >> 16) & 0xFF) + ((tmp >>  8) & 0xFF) + (tmp & 0xFF) - 510;
                        }
		        break;
		case 0:
		        Q += 36;
		        break;
		case 1:
                        bits = get_bits(bs, 1);
			for (k=0; k<36; k+=3)
			{
				unsigned int idx = Huffman_Decode_Q(bs, bits);
				*Q++ = idx30[idx];
				*Q++ = idx31[idx];
				*Q++ = idx32[idx];
			}
		        break;
		case 2:
                        bits = get_bits(bs, 1) + 2;
			for (k=0; k<36; k+=2)
			{
				unsigned int idx = Huffman_Decode_Q(bs, bits);
				*Q++ = idx50[idx];
				*Q++ = idx51[idx];
			}
		        break;
		case 3:
                        bits = get_bits(bs, 1) + 4;
			for (k=0; k<36; ++k) *Q++ = Huffman_Decode_Q(bs, bits);
		        break;
		case 4:
                        bits = get_bits(bs, 1) + 6;
			for (k=0; k<36; ++k) *Q++ = Huffman_Decode_Q(bs, bits);
		        break;
		case 5:
                        bits = get_bits(bs, 1) + 8;
			for (k=0; k<36; ++k) *Q++ = Huffman_Decode_Q(bs, bits);
		        break;
		case 6:
                        bits = get_bits(bs, 1) + 10;
			for (k=0; k<36; ++k) *Q++ = Huffman_Decode_Q(bs, bits);
		        break;
		case 7:
                        bits = get_bits(bs, 1) + 12;
			for (k=0; k<36; ++k) *Q++ = Huffman_Decode_Q(bs, bits);
		        break;
                case 8: case 9: case 10: case 11: case 12: case 13: case 14: case 15: case 16: case 17:
                        bits = mpp->Res[ch][n];
			for (k=0; k<36; ++k) *Q++ = get_bits(bs, bits - 1) - Dc[bits];
		        break;
		case 18:
		        mpp->Res[ch][n] = 0;
		        Q += 36;
		        break;
                default:
                        return;
		}
	}
}

const fixed_t ISMatrix [32] [2] =
{
    {  0x10000000,  0x00000000 }, {  0x0fb14be0,  0x031f1708 },
    {  0x0ec835e0,  0x061f78a8 }, {  0x0d4db310,  0x08e39da0 },
    {  0x0b504f30,  0x0b504f30 }, {  0x08e39da0,  0x0d4db310 },
    {  0x061f78a8,  0x0ec835e0 }, {  0x031f1708,  0x0fb14be0 },
    {  0x00000000,  0x10000000 }, { -0x031f1708,  0x0fb14be0 },
    { -0x061f78a8,  0x0ec835e0 }, { -0x08e39da0,  0x0d4db310 },
    { -0x0b504f30,  0x0b504f30 }, { -0x0d4db310,  0x08e39da0 },
    { -0x0ec835e0,  0x061f78a8 }, { -0x0fb14be0,  0x031f1708 },
    { -0x10000000,  0x00000000 }, { -0x0fb14be0, -0x031f1708 },
    { -0x0ec835e0, -0x061f78a8 }, { -0x0d4db310, -0x08e39da0 },
    { -0x0b504f30, -0x0b504f30 }, { -0x08e39da0, -0x0d4db310 },
    { -0x061f78a8, -0x0ec835e0 }, { -0x031f1708, -0x0fb14be0 },
    {  0x00000000, -0x10000000 }, {  0x031f1708, -0x0fb14be0 },
    {  0x061f78a8, -0x0ec835e0 }, {  0x08e39da0, -0x0d4db310 },
    {  0x0b504f30, -0x0b504f30 }, {  0x0d4db310, -0x08e39da0 },
    {  0x0ec835e0, -0x061f78a8 }, {  0x0fb14be0, -0x031f1708 },
};

void Requantisierung(mpc_params *mpp, int m)
{
    int Band,Last_Band,n;

    //Requantisierung und Skalierung der Subband- und Subframesamples
    Last_Band = mpp->Min_Band-1;

    for (Band=0; Band<=Last_Band; ++Band)
    {
        fixed_t templ,tempr;
        signed int Res_L = mpp->Res[0][Band], Res_R = mpp->Res[1][Band];

        if (mpp->MS_Flag[Band])
        {
            for (n=0; n<12; ++n)
            {
                if ( Res_R < -1 ) {
                    if(Res_L)
                    {
                        f_mul(templ, Requantisieren(mpp->Q[Band][0][12*m+n], Res_L), SCF(mpp->SCF_Index[0][Band][m]))
                        f_mul(mpp->Y[n][0][Band], templ, ISMatrix[mpp->MS_Flag[Band]][0]);
                        f_mul(mpp->Y[n][1][Band], templ, ISMatrix[mpp->MS_Flag[Band]][1]);
                    }
                    else
                    	mpp->Y[n][0][Band] = mpp->Y[n][1][Band] = 0;
                }
                else if ( Res_L < -1 ) {
                    if(Res_R)
                    {
                        f_mul(templ, Requantisieren(mpp->Q[Band][1][12*m+n], Res_R), SCF(mpp->SCF_Index[1][Band][m]))
                        f_mul(mpp->Y[n][1][Band], templ, ISMatrix[mpp->MS_Flag[Band]][0]);
                        f_mul(mpp->Y[n][0][Band], templ, ISMatrix[mpp->MS_Flag[Band]][1]);
                    }
                    else
                        mpp->Y[n][0][Band] = mpp->Y[n][1][Band] = 0;
                }
                else
                {
                    if(Res_L)
                        f_mul(templ, Requantisieren(mpp->Q[Band][0][12*m+n], Res_L), SCF(mpp->SCF_Index[0][Band][m]))
                    else
                        templ = 0;
                    if(Res_R)
                        f_mul(tempr, Requantisieren(mpp->Q[Band][1][12*m+n], Res_R), SCF(mpp->SCF_Index[1][Band][m]))
                    else
                        tempr = 0;
                    mpp->Y[n][0][Band] = (templ+tempr);
                    mpp->Y[n][1][Band] = (templ-tempr);
                }
            }
        }
        else /*if (MS_Flag[Band]==0)*/
        {
            for (n=0; n<12; ++n)
            {
                if(Res_L)
                    f_mul(mpp->Y[n][0][Band], Requantisieren(mpp->Q[Band][0][12*m+n], Res_L), SCF(mpp->SCF_Index[0][Band][m]))
                else
                    mpp->Y[n][0][Band] = 0;
                if(Res_R)
                    f_mul(mpp->Y[n][1][Band], Requantisieren(mpp->Q[Band][1][12*m+n], Res_R), SCF(mpp->SCF_Index[1][Band][m]))
                else
                    mpp->Y[n][1][Band] = 0;
            }
        }
    }
}

void Intensity_Stereo_Decode(mpc_params *mpp, int m)
{
    if (mpp->IS_Flag)
    {
        int Band,n;
        signed char * Res = &(mpp->Res[0][mpp->Min_Band]);
        const fixed_t norm = 0x0d744fcd; /* 1/sqrt(sqrt(2)) */

        //Requantisierung und Wiederherstellung des Stereo-Images
        for (Band=mpp->Min_Band; Band<=mpp->Max_Band; ++Band)
        {
            signed int Res_L = *Res++;

	    if (Res_L)
                for (n=0; n<12; ++n)
		{
                    fixed_t temp;

                    f_mul(temp, Requantisieren(mpp->Q[Band][0][12*m+n], Res_L), norm);
                    f_mul(mpp->Y[n][0][Band], temp, SCF(mpp->SCF_Index[0][Band][m]));
                    f_mul(mpp->Y[n][1][Band], temp, SCF(mpp->SCF_Index[1][Band][m]));
		}
	    else
                for (n=0; n<12; ++n)
                    mpp->Y[n][0][Band] = mpp->Y[n][1][Band] = 0;
        }
    }
}

unsigned long
MPPDecodeData( tMPCInstance *Instance, short *left, short *right, Bit_stream_struc *bs )
{
    short *outsample[2];
    mpc_params *inst = (mpc_params *)Instance;
    Bit_stream_struc StartBS;
    int i;

    if((inst->N >= inst->OverallFrames) && (!inst->TrueGaplessPresent))
        return 3;

    outsample[0] = left;
    outsample[1] = right;

    if ((inst->N % inst->Blockgroesse) == 0)
        inst->CurrBlkSize = get_bits(bs, 20);

    StartBS.bufptr = bs->bufptr;
    StartBS.bitidx = bs->bitidx;

    if (inst->StreamVersion<=6) Lese_Bitstrom_SV6(inst, bs);
    else if (inst->StreamVersion>=7) Lese_Bitstrom_SV7(inst, bs);

    inst->CurrBlkSize -= (bs->bufptr - StartBS.bufptr) * 32 + (bs->bitidx - StartBS.bitidx);
    if(((++inst->N % inst->Blockgroesse) == 0) && inst->CurrBlkSize)
        return 4;

    for(i = 0; i < 3; i++)
    {
    	Requantisierung(inst, i);
        Intensity_Stereo_Decode(inst, i);
        SubBandSynthesis(&inst->ss, inst->Y, outsample, 2, 12);
    }

    if(inst->N == 1)
    {
    	memset(left, 0, DELAY * sizeof(short));
    	memset(right, 0, DELAY * sizeof(short));
    }
    else if(inst->N == inst->OverallFrames)
    {
        if (inst->StreamVersion < 6)
        {
            memset(left, 0, MPC_BLOCK_SIZE * sizeof(short));
    	    memset(right, 0, MPC_BLOCK_SIZE * sizeof(short));
            inst->TrueGaplessPresent = 0;
        }
        else
        {
            int valid = get_bits(bs, 11);

            // Old encoder writes a 0 instead of a 1152, Bugfix
            if (valid == 0) valid = MPC_BLOCK_SIZE;

            valid += DELAY;

            if (valid <= MPC_BLOCK_SIZE)
            {
                memset(left + valid, 0, (MPC_BLOCK_SIZE - valid) * sizeof(short));
    	        memset(right + valid, 0, (MPC_BLOCK_SIZE - valid) * sizeof(short));
                inst->TrueGaplessPresent = 0;
            }
            else
                valid -= MPC_BLOCK_SIZE;

            inst->LastValidSamples = valid;
        }
    }
    else if(inst->N > inst->OverallFrames)
    {
        memset(left + inst->LastValidSamples, 0, (MPC_BLOCK_SIZE - inst->LastValidSamples) * sizeof(short));
    	memset(right + inst->LastValidSamples, 0, (MPC_BLOCK_SIZE - inst->LastValidSamples) * sizeof(short));
        inst->TrueGaplessPresent = 0;
    }

    return 0;
}

const unsigned long sftable[4] = { 44100, 48000, 37800, 32000 };

unsigned long
InitMPPAudio( tMPCInstance *Instance, Bit_stream_struc *bs )
{
    unsigned long Tag;
    int Max_Band_desired;
    mpc_params *inst = (mpc_params *)Instance;

    inst->StreamVersion = 4;
    Tag = preview_bits(bs, 32);
    if((Tag & 0xffffff) == 0x2B504D)
    {
    	inst->StreamVersion = Tag >> 24;
        if (inst->StreamVersion < 7)
	    return 2;
        else if (inst->StreamVersion == 0x17)
            inst->StreamVersion = 7;
    	ignore_bits(bs, 32);
    }

    if (inst->StreamVersion < 7)
    {
	//Auslesen des Fileheaders
	inst->Bitrate  = get_bits(bs, 9);
	inst->IS_Flag  = get_bits(bs, 1);
	inst->MS_used  = get_bits(bs, 1);
	inst->StreamVersion = get_bits(bs, 10);
        if ((inst->StreamVersion<4) || (inst->StreamVersion>6))
	    return 2;
	Max_Band_desired = get_bits(bs, 5);
	inst->Blockgroesse = get_bits(bs, 6);
	if (inst->StreamVersion<5) inst->OverallFrames = get_bits(bs, 16);
	else                       inst->OverallFrames = get_bits(bs, 32);
        inst->SampleRate = 44100;
        inst->TrueGaplessPresent = 0;
        inst->LastValidSamples   = 0;
    }
    else if (inst->StreamVersion == 7)
    {
	//Auslesen des Fileheaders
	inst->Bitrate = 0; inst->Blockgroesse = 1;
	inst->OverallFrames = get_bits(bs, 32);
	inst->IS_Flag  = get_bits(bs, 1);
	inst->MS_used  = get_bits(bs, 1);
	Max_Band_desired = get_bits(bs, 6);

	// reading the profile
	ignore_bits(bs, 4);

	// byte align
	ignore_bits(bs, 2);

        // reading the samplerate
        inst->SampleRate = sftable [get_bits(bs, 2)];

	// reading peak and gain values from the file (or use useful values if they are absent)
	ignore_bits(bs, 16);
	ignore_bits(bs, 16);
	ignore_bits(bs, 16);
	ignore_bits(bs, 16);
	ignore_bits(bs, 16);

        // reading true gapless
        inst->TrueGaplessPresent = get_bits(bs, 1);
        inst->LastValidSamples   = get_bits(bs, 11);
        ignore_bits(bs, 20);

        // reserved bytes for future use
        ignore_bits(bs,  8);
    }
    else
        return 2;

    Synthesis_Init(&inst->ss);
    Init_QuantTab(inst, Max_Band_desired);
    inst->N = 0;
    memset(inst->SCF_Index, 0, sizeof(inst->SCF_Index));
    memset(inst->Y, 0, sizeof(inst->Y));

    return 0;
}

unsigned long MPPGetSamplerate( tMPCInstance *Instance)
{
    return (((mpc_params *)Instance)->SampleRate);
}

unsigned long MPPGetAvgBitrate( tMPCInstance *Instance)
{
    return (((mpc_params *)Instance)->Bitrate);
}

unsigned long MPPGetOverallFrames( tMPCInstance *Instance)
{
    return (((mpc_params *)Instance)->OverallFrames);
}