//****************************************************************************
//
// CODECMPP.C - Codec interface driver for the MPC audio decoder.
//
//****************************************************************************

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "bitstream.h"
#include "mpc.h"

//****************************************************************************
//
// A structure which defines the persistent state of the MPC decoder.
//
//****************************************************************************
typedef struct
{
    //
    // A pointer to the buffer containing the persistent internal state of the
    // MPP decoder library.
    //
    tMPCInstance pMPCInstance;

    //
    // A buffer to contain the encoded MPP audio.
    //
    char pcEncodedData[2048];

    //
    // Buffers to contain the decoded MPP audio.
    //
    short psLeft [MPC_BLOCK_SIZE * 2];
    short psRight[MPC_BLOCK_SIZE * 2];

    //
    // The buffer to which we write decoded MPP data.
    //
    BufferState *pOutput;

    //
    // The MPP bitstream pointer.
    //
    Bit_stream_struc sBS;

    //
    // The length of the encoded MPP file.
    //
    unsigned long ulFileLength;

    //
    // The time length of the file.
    //
    unsigned long ulTimeLength;

    //
    // The number of samples which have been decoded.
    //
    unsigned long ulTimePos;

    //
    // The sum of bit rates for each MPP frame that has been decoded.
    //
    unsigned long ulBitRate;

    //
    // The sample rate of the decoded PCM stream.
    //
    unsigned short ulSampleRate;

    //
    // The number of bytes in pcEncodedData which contain valid encoded MPP
    // data.
    //
    unsigned short ulValid;

} tMPP;

//****************************************************************************
//
// The codec plug-in entry point for the MPP decoder.
//
//****************************************************************************
unsigned long
MPPIoctl(unsigned long ulIoctl, unsigned long ulParam1, unsigned long ulParam2,
         unsigned long ulParam3)
{
    tMPP *pMPP;

    //
    // The first parameter is a pointer to the MPP persistent data.
    //
    pMPP = *((tMPP **)ulParam1);

    //
    // Determine what to do based on the specified IOCTL.
    //
    switch(ulIoctl)
    {
        //
        // Return the name for this codec.
        //
        case IOCTL_CODEC_GETNAME:
        {
            char **ppcName;

            //
            // The secod parameter is a pointer for the name.
            //
            ppcName = (char **)ulParam2;

            //
            // Return the name of this codec.  The first four characters are
            // the short name and the string starting at the fifth character
            // is the long name.
            //
            *ppcName = (char *)"MPC Musepack";

            //
            // Success.
            //
            return(1);
        }

        //
        // Return the bitrate at which this file is encoded.
        //
        case IOCTL_CODEC_GETBITRATE:
        {
            unsigned long *pulBitRate;

            //
            // The second parameter is a pointer for the bitrate.
            //
            pulBitRate = (unsigned long *)ulParam2;

            //
            // Return the average bitrate of the file.
            //
            *pulBitRate = pMPP->ulBitRate;

            //
            // Success.
            //
            return(1);
        }

        //
        // Return the sample rate at which this file is encoded.
        //
        case IOCTL_CODEC_GETSAMPLERATE:
        {
            unsigned long *pulSampleRate;

            //
            // The second parameter is a pointer for the sample rate.
            //
            pulSampleRate = (unsigned long *)ulParam2;

            //
            // Return the sample rate of the file.
            //
            *pulSampleRate = pMPP->ulSampleRate;

            //
            // Success.
            //
            return(1);
        }

        //
        // Return the number of channels in the file.
        //
        case IOCTL_CODEC_GETCHANNELS:
        {
            unsigned long *pulChannels;

            //
            // The second parameter is a pointer for the number of channels.
            //
            pulChannels = (unsigned long *)ulParam2;

            //
            // Return the number of channels in the file.
            //
            *pulChannels = 2;

            //
            // Success.
            //
            return(1);
        }

        //
        // Return the length (in seconds) of the file.
        //
        case IOCTL_CODEC_GETLENGTH:
        {
            unsigned long *pulLength;

            //
            // The second parameter is a pointer for the number of seconds.
            //
            pulLength = (unsigned long *)ulParam2;

            //
            // Return the length of the file.
            //
            *pulLength = pMPP->ulTimeLength;

            //
            // Success.
            //
            return(1);
        }

        //
        // Return the current position (in seconds) within the file.
        //
        case IOCTL_CODEC_GETTIME:
        {
            unsigned long *pulTime;

            //
            // The second parameter is a pointer for the number of seconds.
            //
            pulTime = (unsigned long *)ulParam2;

            //
            // Determine the time based on the sample rate.
            //
            *pulTime = ((pMPP->ulTimePos / pMPP->ulSampleRate) * 1000) +
                       (((pMPP->ulTimePos % pMPP->ulSampleRate) * 1000) / pMPP->ulSampleRate);

            //
            // Success.
            //
            return(1);
        }

        //
        // Prepare the codec to encode or decode a file.
        //
        case IOCTL_CODEC_OPEN:
        {
            FILE *fp = (FILE *)ulParam2;
	    unsigned long ulStart;
            char *ptr;

            //
            // If we are being asked to encode an MPP file, then return a
            // failure since we can only decode MPP.
            //
            if(ulParam3 & CODEC_OPEN_ENCODE)
            {
                //
                // Failure.
                //
                return 0;
            }

            //
            // If file ext is not "mpc" or "mp+", then return a
            // failure since it is not a MPC File.
            //
            ptr = strrchr(fp->FileName, '.');
            if(!ptr || ((ptr[1] != 'M') && (ptr[1] != 'm')) ||
                       ((ptr[2] != 'P') && (ptr[2] != 'p')) ||
                       ((ptr[3] != 'P') && (ptr[3] != 'p') &&
                        (ptr[3] != 'C') && (ptr[3] != 'c') &&
                        (ptr[3] != '+')))
            {
                //
                // Failure.
                //
                return 0;
            }

            //
            // Alloc the MPP persistent state.
            //
            pMPP = (tMPP *)malloc(sizeof(tMPP));
            if(pMPP)
                *((tMPP **)ulParam1) = pMPP;
            else
              	return(0);

            //
            // Get the length of the MPP bitstream.
            //
            pMPP->ulFileLength = FSlength(fp);

            //
            // Read 512 Byte from the file.
            //
            FSseek(fp, 0);
            pMPP->ulValid = FSread(fp, pMPP->pcEncodedData, 512);
            if(pMPP->ulValid != 512)
            {
                free(pMPP);
                return(0);
            }

            //
            // If ID3v2 present, ignore it
            //
            ulStart = ID3v2Jump(pMPP->pcEncodedData);
            if(ulStart)
            {
            	if(FSseek(fp, ulStart) == 0)
                {
                    free(pMPP);
                    return(0);
                }
                pMPP->ulValid = FSread(fp, pMPP->pcEncodedData, 512);
                if(pMPP->ulValid != 512)
                {
                    free(pMPP);
                    return(0);
                }
            }

            //
            // Initialize the bitstream pointer structure.
            //
            pMPP->sBS.bufptr = (unsigned int *)pMPP->pcEncodedData;
            pMPP->sBS.bitidx = 0;

            //
            // Initialize the MPP library.
            //
            if(InitMPPAudio(&pMPP->pMPCInstance, &pMPP->sBS))
            {
                free(pMPP);
                return(0);
            }

            //
            // Get the length of the file.
            //
            pMPP->ulTimeLength = MPPGetOverallFrames(&pMPP->pMPCInstance) * MPC_BLOCK_SIZE / 44100;

            //
            // Get the bit rate and sample rate of the file.
            //
            pMPP->ulSampleRate = MPPGetSamplerate(&pMPP->pMPCInstance);
            pMPP->ulBitRate = MPPGetAvgBitrate(&pMPP->pMPCInstance);
            if(pMPP->ulBitRate == 0)
                pMPP->ulBitRate = pMPP->ulFileLength * 8 /
                                  pMPP->ulTimeLength / 1000;

            //
            // The position is zero.
            //
            pMPP->ulTimePos = 0;

            //
            // Success.
            //
            return 1;
        }

        //
        // Set the output buffer for the decoder.
        //
        case IOCTL_CODEC_SETBUFFER:
        {
            //
            // The second parameter is a pointer to the output buffer.
            //
            pMPP->pOutput = (BufferState *)ulParam2;

            //
            // Provide the output buffer with our data buffers.
            //
            BufferSetBuffer((BufferState *)ulParam2, pMPP->psLeft, pMPP->psRight,
                             MPC_BLOCK_SIZE * 2);

            //
            // Success.
            //
            return 1;
        }

	//
        // Decode a frame of data.
        //
        case IOCTL_CODEC_DECODE:
        {
            short *psLeft, *psRight;
            long lLength;
            FILE * fp;

	    fp = (FILE *)ulParam2;

	    //
            // Make sure the bitstream pointer structure makes sense (i.e. it
            // is pointing to a valid bit in a word, not at the bit following
            // the last bit of the previous word).
            //
            if(pMPP->sBS.bitidx == 32)
            {
                pMPP->sBS.bufptr++;
                pMPP->sBS.bitidx = 0;
            }

            //
            // If the bitstream pointer is not at the beginning of the local
            // bitstream buffer, then copy the remaining data back to the
            // beginning.
            //
            if((int)pMPP->sBS.bufptr != (int)pMPP->pcEncodedData)
            {
                //
                // Update the count of valid bytes in the local bitstream
                // buffer.
                //
                pMPP->ulValid -= (int)pMPP->sBS.bufptr -
                                 (int)pMPP->pcEncodedData;

                //
                // Copy the remainder of the data from the end of the local
                // bitstream buffer to the beginning.
                //
                memcpy(pMPP->pcEncodedData, pMPP->sBS.bufptr, (int)pMPP->ulValid);

                //
                // Update the bitstream pointer structure.
                //
                pMPP->sBS.bufptr = (unsigned int *)pMPP->pcEncodedData;
            }

            //
            // See if there is any additional MPP data that we can copy into
            // the local bitstream buffer.
            //
            if((pMPP->ulValid <= (2048 - 512)) &&
               (FSposition(fp) < pMPP->ulFileLength))
            {
            	//
                // Read MPP data to the end of the local bitstream buffer.
                // Since the ARM MPP decoder needs the bytes within each word
                // swapped, perform the byte swap as we read, and update the
                // count of valid bytes.
                //
                pMPP->ulValid += FSread(fp, pMPP->pcEncodedData + pMPP->ulValid,
                                           (2048 - pMPP->ulValid) & ~511);
            }
            if(pMPP->ulValid < 320)
                return 0;

            //
            // Wait until there is space in the output buffer.  This tells us
            // when it is OK to write more data into the output buffer.
            //
            while(BufferSpaceAvailable(pMPP->pOutput) < MPC_BLOCK_SIZE)
            {
                //
                // Put the EP7209 into IDLE mode until the next interrupt
                // occurs.
                //
                Halt();
            }

            //
            // Get a pointer to the available space in the output buffer.
            //
            BufferGetWritePointer(pMPP->pOutput, &psLeft, &psRight, &lLength);

            if(lLength <= 0)
                return 1;

            //
            // Decode the data frame.
            //
            if(MPPDecodeData(&pMPP->pMPCInstance, psLeft, psRight, &pMPP->sBS))
            {
                //
                // We could not decode the frame, so return an error.
                //
                return 0;
            }

            //
            // Update the output buffer's write pointer.
            //
            BufferUpdateWritePointer(pMPP->pOutput, MPC_BLOCK_SIZE);

            //
            // Increment the time based on the number of samples.
            //
            pMPP->ulTimePos += MPC_BLOCK_SIZE;

            //
            // Success.
            //
            return 1;
        }

        //
        // Seek.
        //
        case IOCTL_CODEC_SEEK:
	{
            //
            // Failure.
            //
            return 0;
	}

        //
        // Encode a frame of data.
        //
        case IOCTL_CODEC_ENCODE:
        {
            //
            // Failure.
            //
            return 0;
        }

        //
        // Cleanup after the codec.
        //
        case IOCTL_CODEC_CLOSE:
        {
            //
            // There's nothing to do, so return success.
            //
            free(pMPP);
            return 1;
        }

        //
        // We should never get here, but just in case...
        //
        default:
        {
            //
            // Return a failure.
            //
            return 0;
        }
    }
}
